/*****************************************************************************
  Exception handlers and startup code for an i.MX6UL target.

  Copyright (c) 2018 by Michael Fischer (www.emb4fun.de).
  Based on the "startup_MCIMX6G3.S" example from Freescale and the
  startup from Rowley, therefore:
  
  Copyright (c) 1997 - 2016 Freescale Semiconductor, Inc.
  Copyright (c) 2014 Rowley Associates Limited.

  This file may be distributed under the terms of the License Agreement
  provided with this software.

  THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING THE
  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *****************************************************************************/

/*****************************************************************************
 *                           Preprocessor Definitions
 *                           ------------------------
 *
 * __NO_SYSTEM_INIT
 *
 *   If defined, the SystemInit() function will not be called. By default
 *   SystemInit() is called after reset to enable the clocks and memories to
 *   be initialised prior to any C startup initialisation.
 *
 *****************************************************************************/

  .section .vectors, "ax"
  .code 32
  .align 0
  .global _vectors
  .global reset_handler
  
/*****************************************************************************
 *                                                                           *
 * Exception Vectors                                                         *
 *                                                                           *
 *****************************************************************************/
_vectors:
  ldr pc, [pc, #reset_handler_address  - . - 8] /* reset */
  ldr pc, [pc, #undef_handler_address  - . - 8] /* undefined instruction */
  ldr pc, [pc, #swi_handler_address    - . - 8] /* swi handler */
  ldr pc, [pc, #pabort_handler_address - . - 8] /* abort prefetch */
  ldr pc, [pc, #dabort_handler_address - . - 8] /* abort data */
  nop
  ldr pc, [pc, #irq_handler_address    - . - 8] /* irq */
  ldr pc, [pc, #fiq_handler_address    - . - 8] /* fiq */

reset_handler_address:
  .word reset_handler
undef_handler_address:
  .word undef_handler
swi_handler_address:
  .word swi_handler
pabort_handler_address:
  .word pabort_handler
dabort_handler_address:
  .word dabort_handler
dummy:
  .word 0   
irq_handler_address:
  .word irq_handler
fiq_handler_address:
  .word fiq_handler


  .section .init, "ax"
  .code 32
  .align 0

/******************************************************************************
 *                                                                            *
 * Default exception handlers                                                 *
 *                                                                            *
 ******************************************************************************/
reset_handler:

  cpsid   i                         /* Mask interrupts */

  /* Reset SCTlr Settings */
  mrc     p15, 0, r0, c1, c0, 0     /* Read CP15 System Control register                  */
  bic     r0,  r0, #(0x1 << 12)     /* Clear I bit 12 to disable I Cache                  */
  bic     r0,  r0, #(0x1 <<  2)     /* Clear C bit  2 to disable D Cache                  */
  bic     r0,  r0, #0x2             /* Clear A bit  1 to disable strict alignment         */
  bic     r0,  r0, #(0x1 << 11)     /* Clear Z bit 11 to disable branch prediction        */
  bic     r0,  r0, #0x1             /* Clear M bit  0 to disable MMU                      */
  mcr     p15, 0, r0, c1, c0, 0     /* Write value back to CP15 System Control register   */

#ifndef __NO_SYSTEM_INIT
  ldr     r0,=SystemInit
  blx     r0
#endif
  cpsie   i                         /* Unmask interrupts             */

  b _start

/******************************************************************************
 *                                                                            *
 * Default exception handlers                                                 *
 * These are declared weak symbols so they can be redefined in user code.     * 
 *                                                                            *
 ******************************************************************************/

undef_handler:
  b .  /* Endless loop */

swi_handler:
  b .  /* Endless loop */

pabort_handler:
  b .  /* Endless loop */

dabort_handler:
  b .  /* Endless loop */
  

    .align 2
    .arm
    .weak irq_handler
    .type irq_handler, %function
irq_handler:
    push    {lr}                         /* Save return address+4                                */
    push    {r0-r3, r12}                 /* Push caller save registers                           */

    MRS     r0, spsr                     /* Save SPRS to allow interrupt reentry                 */
    push    {r0}

    MRC     P15, 4, r1, C15, C0, 0       /* Get GIC base address                                 */
    ADD     r1, r1, #0x2000              /* r1: GICC base address                                */
    LDR     r0, [r1, #0xC]               /* r0: IAR                                              */

    push    {r0, r1}

    CPS     #0x13                        /* Change to Supervisor mode to allow interrupt reentry */

    push    {lr}                         /* Save Supervisor lr                                   */
    LDR     r2, =SystemIrqHandler
    BLX     r2                           /* Call SystemIrqHandler with param GCC                 */
    POP     {lr}

    CPS     #0x12                        /* Back to IRQ mode                                     */

    POP     {r0, r1}

    STR     r0, [r1, #0x10]              /* Now IRQ handler finished: write to EOIR              */

    POP     {r0}
    MSR     spsr_cxsf, r0

    POP     {r0-r3, r12}
    POP     {lr}
    SUBS    pc, lr, #4
    .size irq_handler, . - irq_handler
  

fiq_handler:
  b .  /* Endless loop */

  .weak undef_handler, swi_handler, pabort_handler, dabort_handler, irq_handler, fiq_handler
  
/*** EOF ***/

